import createEmotion from '@emotion/css/create-instance';
import EventEmitter from 'eventemitter3';
import { ContainerEvent } from './event/continer';
import { RenderCode } from './render';
import * as schedule from './scheduler';
import { CodeUrlAdapter } from './handle/adapter';
import { codeUrlManager } from './code-url';

export { createEmotion };
export type RenderData = {
  id: string; // id不会重复
  children?: string[];
  parents?: string[];
  className?: string;
  style?: React.CSSProperties | { [key: string]: string };
  code: RenderCode | string;
  codeId?: string;
  shadowRoot?: boolean;
  hash?: string;
  [key: string]: any;
};

export type ContainerOpts = {
  root?: HTMLDivElement | string;
  shadowRoot?: ShadowRoot;
  data?: RenderData[];
  adapter?: any;
};

export class Container {
  data: RenderData[];
  root: HTMLDivElement;
  globalCss: any;
  event: EventEmitter;
  entryId: string = '';
  adapter: CodeUrlAdapter;
  constructor(opts: ContainerOpts) {
    const data = opts.data || [];
    this.data = data;
    const rootElement = typeof opts.root === 'string' ? document.querySelector(opts.root) : opts.root;
    this.root = rootElement || (document.body as any);
    this.globalCss = createEmotion({ key: 'css-global', speedy: true }).css;
    const event = new EventEmitter();
    this.event = event;
    const listening = this.root.dataset.listening;
    this.adapter = opts.adapter || this.createAdapter();
    if (listening !== 'true') {
      this.root.addEventListener('onContainer', (e: ContainerEvent) => {
        const { type, data } = e.data;
        this.event.emit(type, data);
      });
    }
  }
  protected createAdapter() {
    const adapter = new CodeUrlAdapter({ codeUrlManager });
    this.adapter = adapter;
    return adapter;
  }
  getData({ id }: { id: string }) {
    const data = this.data;
    const node = data.find((node) => node.id === id);
    return node;
  }
  updateData(data: RenderData[]) {
    this.data = this.data.map((node) => {
      const find = data.find((item) => item.id === node.id);
      return find || node;
    });
  }
  /**
   * 渲染模块
   * @param cid
   * @param parentElement
   * @param pid
   * @returns
   */
  renderChildren(cid: string, parentElement?: any, pid?: any, isNew = true): void | HTMLDivElement {
    const data = this.data;
    const globalCss = this.globalCss;
    if (!parentElement) {
      parentElement = this.root;
      this.root.dataset.entryId = cid;
      this.entryId = cid;
    }
    const renderChildren = this.renderChildren.bind(this);
    const node = data.find((node: RenderData) => node.id === cid);
    if (!node) {
      console.warn('node not found', cid);
      return;
    }
    let el;
    if (isNew) {
      el = document.createElement('div');
      parentElement.appendChild(el);
      const parentIds = parentElement.dataset.parentIds || '';

      el.dataset.cid = cid;
      el.dataset.pid = pid;
      el.dataset.parentIds = parentIds ? parentIds + ',' + cid : cid;
      // 代码渲染
      const { renderFn } = this.getContext({ renderEl: el, entryId: cid });
      schedule.addTask(renderFn, 'high');
      // 代码渲染结束
    } else {
      el = parentElement.querySelector(`[data-cid="${cid}"]`);
    }

    if (!el) {
      console.warn('when no createNewElement el not found', cid);
      return;
    }
    const childrenIds = node.children || [];
    childrenIds.forEach((childId) => {
      renderChildren(childId, el, cid);
    });
    return el;
  }
  /**
   * 渲染根节点
   * @param entryId
   * @returns
   */
  async renderRoot(entryId: string) {
    if (this.entryId && this.entryId !== entryId) {
      await this.destroy();
    } else if (this.entryId) {
      await this.destroy(entryId);
    }
    this.renderChildren(entryId);
  }
  async reRender() {
    await this.destroy();
    this.renderChildren(this.entryId);
  }
  protected getContext({ renderEl, entryId }: { renderEl: HTMLDivElement; entryId: string }) {
    const node = this.data.find((node) => node.id === entryId);
    const { style, code } = node;
    const event = this.event;
    const globalCss = this.globalCss;
    const root = renderEl;
    const adapter = this.adapter;
    const containerThis = this;
    const shadowRoot = node.shadowRoot ? renderEl.attachShadow({ mode: 'open' }) : null;
    const { css, sheet, cache } = createEmotion({ key: 'css' });
    if (shadowRoot) {
      cache.insert = (selector: string, serialized: any, sheet: any, shouldCache: boolean) => {
        const style = document.createElement('style');
        style.textContent = selector;
        shadowRoot.appendChild(style);
      };
    }
    const cid = entryId;
    renderEl.className = '';
    if (renderEl.style) {
      renderEl.className = !shadowRoot ? globalCss(style as any) : css(style as any);
    }
    renderEl.classList.add(cid, 'kv-container');
    if (shadowRoot) {
      const slot = document.createElement('slot');
      shadowRoot.appendChild(slot);
    }
    const renderFn = async () => {
      const { code: newCode } = await adapter.handle(node as any);
      const { render, unmount } = newCode || {};
      let renderRoot = node.shadowRoot ? shadowRoot : root;
      const ctx = {
        root: root,
        shadowRoot,
        renderRoot,
        event,
        container: containerThis,
        code: newCode as RenderCode,
        data: node,
        css: shadowRoot ? css : globalCss,
      };
      if (render) {
        render(ctx);
      } else {
        // no render, so no insert
      }
      if (event) {
        const destroy = (id: string) => {
          if (id) {
            if (id === cid || node?.parents?.find?.((item) => item === id)) {
              unmount?.(ctx); // 销毁父亲元素有这个元素的
              event.off('destroy', destroy); // 移除监听
            } else {
              // console.warn('destroy id not found, and not find parentIds', id, 'currentid', cid);
            }
          } else if (!id) {
            unmount?.(ctx); //  所有的都销毁
            event.off('destroy', destroy);
          }
          // 不需要销毁子元素
        };
        event.on('destroy', destroy);
      }
    };
    return { renderFn, shadowRoot };
  }
  // 更新渲染
  async render(entryId: string, el?: HTMLDivElement) {
    if (schedule.getLoading()) {
      console.log('schedule is loading');
    }
    if (!this.entryId) {
      console.warn('root is use renderRoot method, not render');
      return this.renderRoot(entryId);
    }
    const isRoot = this.entryId === entryId;
    if (isRoot) {
      await this.reRender();
      return;
    }
    let renderEl = el;
    if (!el) {
      renderEl = this.root.querySelector(`[data-cid="${entryId}"]`);
    }
    if (!renderEl) {
      console.warn('render element not found', entryId);
      return;
    } else {
      await this.destroy(entryId);
    }
    const parentElement = renderEl.parentElement;
    const pid = renderEl.dataset.pid;
    const { renderFn } = this.getContext({ renderEl, entryId });
    schedule.addTask(renderFn, 'high');
    this.renderChildren(entryId, parentElement, pid, false);
  }

  async destroy(id?: string) {
    if (!id) {
      this.root.innerHTML = '';
    } else {
      const elements = this.root.querySelectorAll(`[data-cid="${id}"]`);
      elements.forEach((element) => {
        element.innerHTML = '';
      });
    }
    this.event.emit('destroy', id);
    await new Promise((resolve) => {
      setTimeout(() => {
        resolve(null);
      }, 200);
    });
  }
  close() {
    this.destroy();
    const event = this.event;
    event && event.removeAllListeners();
  }
}

export class ContainerOne extends Container {
  constructor(opts: ContainerOpts) {
    super(opts);
  }
  async renderOne({ code: RenderCode }) {
    const newData = { codeId: 'test-reneder-one-code-id', id: 'test-render-one', code: RenderCode };
    this.data = [newData];
    setTimeout(() => {
      this.renderRoot('test-render-one');
    }, 1000);
  }
}
export const mount = ({ render, unmount }, root: string | HTMLDivElement) => {
  let _root = typeof root === 'string' ? document.querySelector(root) : root;
  _root = _root || (document.body as any);
  render({ root: _root });
};
